from typing import FrozenSet

from .yaku import Yaku

richi = Yaku("richi", 1, 1)
ippatsu = Yaku("ippatsu", 1, 1)
rinshan = Yaku("rinshan", 1, 0)
chankan = Yaku("chankan", 1, 0)
haitei = Yaku("haitei", 1, 0)
houtei = Yaku("houtei", 1, 0)
w_richi = Yaku("w_richi", 2, 2)
tenhou = Yaku("tenhou", 13, 13, True)
chihou = Yaku("chihou", 13, 13, True)

all_extra_yaku: FrozenSet[Yaku] = frozenset({
    richi, ippatsu, rinshan, chankan, haitei, houtei,
    w_richi,
    tenhou, chihou
})

__all__ = (
    "all_extra_yaku",
    "richi", "ippatsu", "rinshan", "chankan", "haitei", "houtei",
    "w_richi",
    "tenhou", "chihou"
)



from mahjong_utils.yaku.yaku import Yaku
from mahjong_utils.yaku.common import all_common_yaku
from mahjong_utils.yaku.extra import all_extra_yaku
from mahjong_utils.yaku.yakuman import all_yakuman

all_yaku = all_common_yaku | all_extra_yaku | all_yakuman

_all_yaku_mapping = dict(map(lambda x: (x.name, x), all_yaku))


def get_yaku(name: str) -> Yaku:
    yaku = _all_yaku_mapping.get(name, None)
    if yaku is None:
        raise ValueError(f"{name} is not a yaku")
    return yaku


__all__ = ("all_yaku", "get_yaku")



from typing import Sequence, Generic, TypeVar, Type

from pydantic.fields import Field
from pydantic.generics import GenericModel
from pydantic.main import BaseModel

from mahjong_utils.models.furo import Furo, Kan
from mahjong_utils.models.hand_pattern import HandPattern, CommonHandPattern
from mahjong_utils.models.tile import Tile

T_HandPattern = TypeVar("T_HandPattern", bound=CommonHandPattern)


class Hand(GenericModel, Generic[T_HandPattern]):
    tiles: Sequence[Tile]
    furo: Sequence[Furo] = Field(default_factory=tuple)
    patterns: Sequence[T_HandPattern]

    @property
    def menzen(self) -> bool:
        for fr in self.furo:
            if not isinstance(fr, Kan) or not fr.ankan:
                return False
        return True

    def __encode__(self) -> dict:
        return dict(
            tiles=[t.__encode__() for t in self.tiles],
            furo=[fr.__encode__() for fr in self.furo],
            patterns=[p.__encode__() for p in self.patterns]
        )

    @classmethod
    def __decode__(cls, data: dict, hand_pattern_type: Type[T_HandPattern] = CommonHandPattern) -> "Hand[T_HandPattern]":
        return Hand(
            tiles=[Tile.__decode__(x) for x in data["tiles"]],
            furo=[Furo.__decode__(x) for x in data["furo"]],
            patterns=[hand_pattern_type.__decode__(x) for x in data["patterns"]]
        )



from abc import ABC, abstractmethod
from typing import Optional, FrozenSet, Tuple
from typing import Set

from pydantic import BaseModel
from pydantic import Field
from stringcase import pascalcase, snakecase

from mahjong_utils.models.hand_pattern import RegularHandPattern, HandPattern, _BaseRegularHandPattern, \
    _BaseChitoiHandPattern, _BaseKokushiHandPattern
from mahjong_utils.models.tatsu import Tatsu
from mahjong_utils.models.tile import Tile, all_yaochu
from mahjong_utils.models.wind import Wind
from mahjong_utils.point_by_han_hu.models import ParentPoint, ChildPoint
from mahjong_utils.yaku import Yaku, get_yaku


class HoraHandPattern(HandPattern, ABC):
    agari: Tile
    tsumo: bool
    hu: int
    self_wind: Optional[Wind]
    round_wind: Optional[Wind]

    @abstractmethod
    def __encode__(self) -> dict:
        raise NotImplementedError()

    @classmethod
    def __decode__(cls, data: dict) -> "HoraHandPattern":
        if data['type'] == 'RegularHoraHandPattern':
            return RegularHoraHandPattern.__decode__(data)
        elif data['type'] == 'ChitoiHoraHandPattern':
            return ChitoiHoraHandPattern.__decode__(data)
        elif data['type'] == 'KokushiHoraHandPattern':
            return KokushiHoraHandPattern.__decode__(data)
        else:
            raise ValueError("invalid type: " + data['type'])


class RegularHoraHandPattern(HoraHandPattern, _BaseRegularHandPattern):
    agari_tatsu: Optional[Tatsu]
    jyantou: Tile

    def __encode__(self) -> dict:
        return dict(
            type="RegularHoraHandPattern",
            agari=self.agari.__encode__(),
            tsumo=self.tsumo,
            self_wind=pascalcase(self.self_wind.name) if self.self_wind is not None else None,
            round_wind=pascalcase(self.round_wind.name) if self.round_wind is not None else None,
            agari_tatsu=self.agari_tatsu.__encode__() if self.agari_tatsu is not None else None,
            pattern=RegularHandPattern(**self.dict()).__encode__()
        )

    @classmethod
    def __decode__(cls, data: dict) -> "RegularHoraHandPattern":
        return RegularHoraHandPattern(
            agari=Tile.__decode__(data["agari"]),
            tsumo=data["tsumo"],
            self_wind=Wind[snakecase(data["selfWind"])] if data["selfWind"] is not None else None,
            round_wind=Wind[snakecase(data["roundWind"])] if data["roundWind"] is not None else None,
            agari_tatsu=Tatsu.__decode__(data["agariTatsu"]) if data["agariTatsu"] is not None else None,
            hu=data["hu"],
            **RegularHandPattern.__decode__(data["pattern"]).dict()
        )


class ChitoiHoraHandPattern(HoraHandPattern, _BaseChitoiHandPattern):
    hu: int = 25
    remaining: Tuple[Tile, ...] = Field(default_factory=tuple)

    def __encode__(self) -> dict:
        return dict(
            type="ChitoiHoraHandPattern",
            pairs=[t.__encode__() for t in self.pairs],
            agari=self.agari.__encode__(),
            tsumo=self.tsumo,
            self_wind=pascalcase(self.self_wind.name) if self.self_wind is not None else None,
            round_wind=pascalcase(self.round_wind.name) if self.round_wind is not None else None,
        )

    @classmethod
    def __decode__(cls, data: dict) -> "ChitoiHoraHandPattern":
        return ChitoiHoraHandPattern(
            agari=Tile.__decode__(data["agari"]),
            pairs=frozenset(Tile.__decode__(t) for t in data["pairs"]),
            tsumo=data["tsumo"],
            self_wind=Wind[snakecase(data["selfWind"])] if data["selfWind"] is not None else None,
            round_wind=Wind[snakecase(data["roundWind"])] if data["roundWind"] is not None else None,
        )


class KokushiHoraHandPattern(HoraHandPattern, _BaseKokushiHandPattern):
    hu: int = 20
    repeated: Tile
    yaochu: FrozenSet[Tile] = all_yaochu
    remaining: Tuple[Tile, ...] = Field(default_factory=tuple)

    @property
    def thirteen_waiting(self) -> bool:
        return self.agari == self.repeated

    def __encode__(self) -> dict:
        return dict(
            type="KokushiHoraHandPattern",
            repeated=self.repeated.__encode__(),
            agari=self.agari.__encode__(),
            tsumo=self.tsumo,
            self_wind=pascalcase(self.self_wind.name) if self.self_wind is not None else None,
            round_wind=pascalcase(self.round_wind.name) if self.round_wind is not None else None,
        )

    @classmethod
    def __decode__(cls, data: dict) -> "KokushiHoraHandPattern":
        return KokushiHoraHandPattern(
            repeated=Tile.__decode__(data["repeated"]),
            agari=Tile.__decode__(data["agari"]),
            tsumo=data["tsumo"],
            self_wind=Wind[snakecase(data["selfWind"])] if data["selfWind"] is not None else None,
            round_wind=Wind[snakecase(data["roundWind"])] if data["roundWind"] is not None else None,
        )


class Hora(BaseModel):
    pattern: HoraHandPattern
    han: int
    dora: int
    yaku: Set[Yaku]
    extra_yaku: Set[Yaku]
    has_yakuman: bool
    parent_point: ParentPoint
    child_point: ChildPoint

    @classmethod
    def __decode__(cls, data: dict) -> "Hora":
        return Hora(
            pattern=HoraHandPattern.__decode__(data["pattern"]),
            han=data["han"],
            dora=data["dora"],
            yaku=set(get_yaku(snakecase(yk)) for yk in data["yaku"]),
            extra_yaku=set(get_yaku(snakecase(yk)) for yk in data["extraYaku"]),
            has_yakuman=data["hasYakuman"],
            parent_point=ParentPoint.__decode__(data["parentPoint"]),
            child_point=ChildPoint.__decode__(data["childPoint"]),
        )

    @property
    def hu(self) -> int:
        return self.pattern.hu

    @property
    def tsumo(self) -> bool:
        return self.pattern.tsumo

    @property
    def self_wind(self) -> Optional[Wind]:
        return self.pattern.self_wind

    @property
    def round_wind(self) -> Optional[Wind]:
        return self.pattern.round_wind

    @property
    def agari(self) -> Tile:
        return self.pattern.agari



import atexit

from .lib import LibMahjongUtils
from .protocol import MahjongUtilsBridge
from .webapi_jar import WebApiJarMahjongUtils, mahjongutils_webapi_jar_path


class RouterMahjongUtils(MahjongUtilsBridge):
    def __init__(self):
        self._delegate = None

    @property
    def delegate(self) -> MahjongUtilsBridge:
        if self._delegate is None:
            if mahjongutils_webapi_jar_path().exists():
                self._delegate = WebApiJarMahjongUtils()
            else:
                self._delegate = LibMahjongUtils()
        return self._delegate

    def call(self, name: str, params: dict) -> dict:
        return self.delegate.call(name, params)

    def close(self):
        return self.delegate.close()


bridge_mahjongutils = RouterMahjongUtils()
atexit.register(bridge_mahjongutils.close)



from abc import abstractmethod, ABC
from typing import Iterable, Optional, Tuple, FrozenSet

from pydantic import Field
from pydantic.main import BaseModel
from typing_extensions import Self

from mahjong_utils.models.furo import Furo, Kan
from mahjong_utils.models.mentsu import Mentsu, Shuntsu, Kotsu
from mahjong_utils.models.tatsu import Tatsu
from mahjong_utils.models.tile import Tile


class HandPattern(BaseModel, ABC):
    @property
    @abstractmethod
    def menzen(self) -> bool:
        raise NotImplementedError()

    @property
    @abstractmethod
    def tiles(self) -> Iterable[Tile]:
        raise NotImplementedError()

    def __hash__(self):
        return hash(self.__class__) + hash(tuple(self.__dict__.values()))


class CommonHandPattern(HandPattern, ABC):
    @abstractmethod
    def __encode__(self) -> dict:
        raise NotImplementedError()

    @classmethod
    def __decode__(cls, data: dict) -> Self:
        if data['type'] == 'RegularHandPattern':
            return RegularHandPattern.__decode__(data)
        elif data['type'] == 'ChitoiHandPattern':
            return ChitoiHandPattern.__decode__(data)
        elif data['type'] == 'KokushiHandPattern':
            return KokushiHandPattern.__decode__(data)
        else:
            raise ValueError("invalid type: " + data['type'])


class _BaseRegularHandPattern(HandPattern):
    """
    表示一个结构分析后的以标准形为目标的手牌
    """

    k: int
    jyantou: Optional[Tile] = None
    menzen_mentsu: Tuple[Mentsu, ...] = Field(default_factory=tuple)
    furo: Tuple[Furo, ...] = Field(default_factory=tuple)
    tatsu: Tuple[Tatsu, ...] = Field(default_factory=tuple)
    remaining: Tuple[Tile, ...] = Field(default_factory=tuple)

    @property
    def mentsu(self) -> Iterable[Mentsu]:
        for mt in self.menzen_mentsu:
            yield mt
        for fr in self.furo:
            yield fr

    @property
    def shuntsu(self) -> Iterable[Shuntsu]:
        for mt in self.mentsu:
            if isinstance(mt, Shuntsu):
                yield mt

    @property
    def kotsu(self) -> Iterable[Kotsu]:
        for mt in self.mentsu:
            if isinstance(mt, Kotsu):
                yield mt

    @property
    def anko(self) -> Iterable[Kotsu]:
        for mt in self.menzen_mentsu:
            if isinstance(mt, Kotsu):
                yield mt
        for fr in self.furo:
            if isinstance(fr, Kan) and fr.ankan:
                yield fr

    @property
    def menzen(self) -> bool:
        for fr in self.furo:
            if not isinstance(fr, Kan) or not fr.ankan:
                return False
        return True

    @property
    def tiles(self) -> Iterable[Tile]:
        if self.jyantou is not None:
            yield self.jyantou
            yield self.jyantou
        for mt in self.mentsu:
            for t in mt.tiles:
                yield t
        for tt in self.tatsu:
            yield tt.first
            yield tt.second
        for t in self.remaining:
            yield t


class RegularHandPattern(_BaseRegularHandPattern, CommonHandPattern):
    def __encode__(self) -> dict:
        return dict(
            type="RegularHandPattern",
            k=self.k,
            jyantou=self.jyantou.__encode__() if self.jyantou is not None else None,
            menzenMentsu=[mt.__encode__() for mt in self.menzen_mentsu],
            furo=[fr.__encode__() for fr in self.furo],
            tatsu=[tt.__encode__() for tt in self.tatsu],
            remaining=[t.__encode__() for t in self.remaining],
        )

    @classmethod
    def __decode__(cls, data: dict) -> Self:
        return RegularHandPattern(
            k=data["k"],
            jyantou=Tile.__decode__(data["jyantou"]) if data["jyantou"] is not None else None,
            menzen_mentsu=tuple(Mentsu.__decode__(x) for x in data["menzenMentsu"]),
            furo=tuple(Furo.__decode__(x) for x in data["furo"]),
            tatsu=tuple(Tatsu.__decode__(x) for x in data["tatsu"]),
            remaining=tuple(Tile.__decode__(x) for x in data["remaining"])
        )


class _BaseChitoiHandPattern(HandPattern):
    """
    表示一个结构分析后的以七对子为目标的手牌
    """

    pairs: FrozenSet[Tile] = Field(default_factory=frozenset)
    remaining: Tuple[Tile, ...] = Field(default_factory=tuple)

    @property
    def menzen(self) -> bool:
        return True

    @property
    def tiles(self) -> Iterable[Tile]:
        for p in self.pairs:
            yield p
            yield p
        for t in self.remaining:
            yield t


class ChitoiHandPattern(_BaseChitoiHandPattern, CommonHandPattern):
    def __encode__(self) -> dict:
        return dict(
            type="ChitoiHandPattern",
            pairs=[t.__encode__() for t in self.pairs],
            remaining=[t.__encode__() for t in self.remaining],
        )

    @classmethod
    def __decode__(cls, data: dict) -> Self:
        return ChitoiHandPattern(
            pairs=frozenset(Tile.__decode__(x) for x in data["pairs"]),
            remaining=tuple(Tile.__decode__(x) for x in data["remaining"])
        )


class _BaseKokushiHandPattern(HandPattern):
    """
    表示一个结构分析后的以国士无双为目标的手牌
    """

    yaochu: FrozenSet[Tile] = Field(default_factory=frozenset)
    repeated: Optional[Tile] = None
    remaining: Tuple[Tile, ...] = Field(default_factory=tuple)

    @property
    def menzen(self) -> bool:
        return True

    @property
    def tiles(self) -> Iterable[Tile]:
        for t in self.yaochu:
            if self.repeated == t:
                yield t
            yield t
        for t in self.remaining:
            yield t


class KokushiHandPattern(_BaseKokushiHandPattern, CommonHandPattern):
    def __encode__(self) -> dict:
        return dict(
            type="KokushiHandPattern",
            yaochu=[t.__encode__() for t in self.yaochu],
            repeated=self.repeated.__encode__() if self.repeated is not None else None,
            remaining=[t.__encode__() for t in self.remaining],
        )

    @classmethod
    def __decode__(cls, data: dict) -> Self:
        return KokushiHandPattern(
            yaochu=frozenset(Tile.__decode__(x) for x in data["yaochu"]),
            repeated=Tile.__decode__(data["repeated"]) if data["repeated"] is not None else None,
            remaining=tuple(Tile.__decode__(x) for x in data["remaining"])
        )


__all__ = ("HandPattern", "CommonHandPattern", "RegularHandPattern", "ChitoiHandPattern", "KokushiHandPattern")



from typing import NamedTuple


class ParentPoint(NamedTuple):
    ron: int
    tsumo: int

    @classmethod
    def __decode__(cls, data: dict) -> "ParentPoint":
        return ParentPoint(
            ron=data["ron"],
            tsumo=data["tsumo"],
        )


class ChildPoint(NamedTuple):
    ron: int
    tsumo_parent: int
    tsumo_child: int

    @classmethod
    def __decode__(cls, data: dict) -> "ChildPoint":
        return ChildPoint(
            ron=data["ron"],
            tsumo_parent=data["tsumoParent"],
            tsumo_child=data["tsumoChild"],
        )



from abc import ABC, abstractmethod
from typing import Optional, Set, Dict, Generic, TypeVar, List, cast

from pydantic import BaseModel
from pydantic.generics import GenericModel
from typing_extensions import Self

from mahjong_utils.models.hand import Hand, T_HandPattern
from mahjong_utils.models.hand_pattern import RegularHandPattern, ChitoiHandPattern, KokushiHandPattern, HandPattern
from mahjong_utils.models.tatsu import Tatsu
from mahjong_utils.models.tile import Tile


class Shanten(BaseModel, ABC):
    shanten: int

    @abstractmethod
    def __encode__(self) -> dict:
        raise NotImplementedError()

    @classmethod
    def __decode__(cls, data: dict) -> "Shanten":
        if data['type'] == 'ShantenWithFuroChance':
            return ShantenWithFuroChance.__decode__(data)
        else:
            return CommonShanten.__decode__(data)


class CommonShanten(Shanten, ABC):
    @classmethod
    def __decode__(cls, data: dict) -> "Shanten":
        if data['type'] == 'ShantenWithoutGot':
            return ShantenWithoutGot.__decode__(data)
        elif data['type'] == 'ShantenWithGot':
            return ShantenWithGot.__decode__(data)
        else:
            raise ValueError("invalid type: " + data['type'])


class Improvement(BaseModel):
    discard: Tile
    advance: Set[Tile]
    advance_num: int

    def __encode__(self) -> dict:
        return dict(
            discard=self.discard.__encode__(),
            advance=[t.__encode__() for t in self.advance],
            advanceNum=self.advance_num
        )

    @classmethod
    def __decode__(cls, data: dict) -> "Improvement":
        return Improvement(
            discard=Tile.__decode__(data["discard"]),
            advance=set(Tile.__decode__(x) for x in data["advance"]),
            advance_num=data["advanceNum"]
        )


def _encode_improvement_dict(improvement: Dict[Tile, List[Improvement]]):
    d = dict()
    for t in improvement:
        d[t.__encode__()] = [x.__encode__() for x in improvement[t]]
    return d


def _decode_improvement_dict(improvement: dict) -> Dict[Tile, List[Improvement]]:
    d = dict()
    for t in improvement:
        d[Tile.__decode__(t)] = [Improvement.__decode__(x) for x in improvement[t]]
    return d


class ShantenWithoutGot(CommonShanten):
    advance: Set[Tile]
    advance_num: int
    good_shape_advance: Optional[Set[Tile]]
    good_shape_advance_num: Optional[int]
    improvement: Optional[Dict[Tile, List[Improvement]]]
    improvement_num: Optional[int]
    good_shape_improvement: Optional[Dict[Tile, List[Improvement]]]
    good_shape_improvement_num: Optional[int]

    def __encode__(self) -> dict:
        return dict(
            type="ShantenWithoutGot",
            shantenNum=self.shanten,
            advance=[t.__encode__() for t in self.advance],
            advanceNum=self.advance_num,
            goodShapeAdvance=[t.__encode__() for t in self.good_shape_advance]
            if self.good_shape_advance is not None else None,
            goodShapeAdvanceNum=self.good_shape_advance_num,
            improvement=_encode_improvement_dict(self.improvement)
            if self.improvement is not None else None,
            improvementNum=self.improvement_num,
            goodShapeImprovement=_encode_improvement_dict(self.good_shape_improvement)
            if self.good_shape_improvement is not None else None,
            goodShapeImprovementNum=self.good_shape_improvement_num,
        )

    @classmethod
    def __decode__(cls, data: dict) -> "ShantenWithoutGot":
        return ShantenWithoutGot(
            shanten=data["shantenNum"],
            advance=set(Tile.__decode__(x) for x in data["advance"]),
            advance_num=data["advanceNum"],
            good_shape_advance=set(Tile.__decode__(x) for x in data["goodShapeAdvance"])
            if data["goodShapeAdvance"] is not None else None,
            good_shape_advance_num=data["goodShapeAdvanceNum"]
            if data["goodShapeAdvanceNum"] is not None else None,
            improvement=_decode_improvement_dict(data["improvement"])
            if data["improvement"] is not None else None,
            improvement_num=data["improvementNum"]
            if data["improvementNum"] is not None else None,
            good_shape_improvement=_decode_improvement_dict(data["goodShapeImprovement"])
            if data["goodShapeImprovement"] is not None else None,
            good_shape_improvement_num=data["goodShapeImprovementNum"]
            if data["goodShapeImprovementNum"] is not None else None,
        )


class ShantenWithGot(CommonShanten):
    discard_to_advance: Dict[Tile, ShantenWithoutGot]
    ankan_to_advance: Dict[Tile, ShantenWithoutGot]

    def __encode__(self) -> dict:
        return dict(
            type="ShantenWithGot",
            shantenNum=self.shanten,
            discardToAdvance=dict((k.__encode__(), v.__encode__()) for (k, v) in self.discard_to_advance.items()),
            ankanToAdvance=dict((k.__encode__(), v.__encode__()) for (k, v) in self.ankan_to_advance.items())
        )

    @classmethod
    def __decode__(cls, data: dict) -> "ShantenWithGot":
        return ShantenWithGot(
            shanten=data["shantenNum"],
            discard_to_advance=dict(
                (Tile.__decode__(k), ShantenWithoutGot.__decode__(v))
                for (k, v) in data["discardToAdvance"].items()),
            ankan_to_advance=dict(
                (Tile.__decode__(k), ShantenWithoutGot.__decode__(v))
                for (k, v) in data["ankanToAdvance"].items())
        )


class ShantenWithFuroChance(Shanten):
    pass_: Optional[ShantenWithoutGot]
    chi: Dict[Tatsu, ShantenWithGot]
    pon: Optional[ShantenWithGot]
    minkan: Optional[ShantenWithoutGot]

    def __encode__(self) -> dict:
        return {
            'type': "ShantenWithFuroChance",
            'shantenNum': self.shanten,
            'pass': self.pass_.__encode__()
            if self.pass_ is not None else None,
            'chi': dict((k.__encode__(), v.__encode__()) for (k, v) in self.chi.items()),
            'pon': self.pon.__encode__()
            if self.pon is not None else None,
            'minkan': self.minkan.__encode__()
            if self.minkan is not None else None
        }

    @classmethod
    def __decode__(cls, data: dict) -> "ShantenWithFuroChance":
        return ShantenWithFuroChance(
            shanten=data["shantenNum"],
            pass_=ShantenWithoutGot.__decode__(data["pass"])
            if data["pass"] is not None else None,
            chi=dict((Tatsu.__decode__(k), ShantenWithGot.__decode__(v)) for (k, v) in data["chi"].items()),
            pon=ShantenWithGot.__decode__(data["pon"])
            if data["pon"] is not None else None,
            minkan=ShantenWithoutGot.__decode__(data["minkan"])
            if data["minkan"] is not None else None,
        )


T_Shanten = TypeVar('T_Shanten', bound=Shanten)


class ShantenResult(GenericModel, Generic[T_Shanten, T_HandPattern], ABC):
    hand: Hand[T_HandPattern]
    shanten_info: T_Shanten

    @abstractmethod
    def __encode__(self) -> dict:
        raise NotImplementedError()

    @classmethod
    def __decode__(cls, data: dict) -> Self:
        if data['type'] == 'FuroChanceShantenResult':
            return FuroChanceShantenResult.__decode__(data)
        else:
            return CommonShantenResult.__decode__(data)

    @property
    def shanten(self) -> int:
        return self.shanten_info.shanten


class CommonShantenResult(ShantenResult[CommonShanten, T_HandPattern], GenericModel, Generic[T_HandPattern], ABC):
    @property
    def advance(self) -> Optional[Set[Tile]]:
        return getattr(self.shanten_info, "advance", None)

    @property
    def advance_num(self) -> Optional[int]:
        return getattr(self.shanten_info, "advance_num", None)

    @property
    def good_shape_advance(self) -> Optional[Set[Tile]]:
        return getattr(self.shanten_info, "good_shape_advance", None)

    @property
    def good_shape_advance_num(self) -> Optional[int]:
        return getattr(self.shanten_info, "good_shape_advance_num", None)

    @property
    def improvement(self) -> Optional[Dict[Tile, List[Improvement]]]:
        return getattr(self.shanten_info, "improvement", None)

    @property
    def improvement_num(self) -> Optional[int]:
        return getattr(self.shanten_info, "improvement_num", None)

    @property
    def good_shape_improvement(self) -> Optional[Dict[Tile, List[Improvement]]]:
        return getattr(self.shanten_info, "good_shape_improvement", None)

    @property
    def good_shape_improvement_num(self) -> Optional[int]:
        return getattr(self.shanten_info, "good_shape_improvement_num", None)

    @property
    def discard_to_advance(self) -> Optional[Dict[Tile, ShantenWithoutGot]]:
        return getattr(self.shanten_info, "discard_to_advance", None)

    @property
    def ankan_to_advance(self) -> Optional[Dict[Tile, ShantenWithoutGot]]:
        return getattr(self.shanten_info, "ankan_to_advance", None)

    @property
    def with_got(self) -> bool:
        return isinstance(self.shanten_info, ShantenWithGot)

    @classmethod
    def __decode__(cls, data: dict) -> Self:
        if data['type'] == 'RegularShantenResult':
            return RegularShantenResult.__decode__(data)
        elif data['type'] == 'ChitoiShantenResult':
            return ChitoiShantenResult.__decode__(data)
        elif data['type'] == 'KokushiShantenResult':
            return KokushiShantenResult.__decode__(data)
        elif data['type'] == 'UnionShantenResult':
            return UnionShantenResult.__decode__(data)
        else:
            raise ValueError("invalid type: " + data['type'])


class RegularShantenResult(CommonShantenResult[RegularHandPattern]):
    def __encode__(self) -> dict:
        return dict(
            type='RegularShantenResult',
            hand=self.hand.__encode__(),
            shantenInfo=self.shanten_info.__encode__()
        )

    @classmethod
    def __decode__(cls, data: dict) -> Self:
        return RegularShantenResult(
            hand=Hand.__decode__(data["hand"], RegularHandPattern),
            shanten_info=cast(CommonShanten, Shanten.__decode__(data["shantenInfo"]))
        )


class ChitoiShantenResult(CommonShantenResult[ChitoiHandPattern]):
    def __encode__(self) -> dict:
        return dict(
            type='ChitoiShantenResult',
            hand=self.hand.__encode__(),
            shantenInfo=self.shanten_info.__encode__()
        )

    @classmethod
    def __decode__(cls, data: dict) -> Self:
        return ChitoiShantenResult(
            hand=Hand.__decode__(data["hand"], ChitoiHandPattern),
            shanten_info=cast(CommonShanten, Shanten.__decode__(data["shantenInfo"]))
        )


class KokushiShantenResult(CommonShantenResult[KokushiHandPattern]):
    def __encode__(self) -> dict:
        return dict(
            type='KokushiShantenResult',
            hand=self.hand.__encode__(),
            shantenInfo=self.shanten_info.__encode__()
        )

    @classmethod
    def __decode__(cls, data: dict) -> Self:
        return KokushiShantenResult(
            hand=Hand.__decode__(data["hand"], KokushiHandPattern),
            shanten_info=cast(CommonShanten, Shanten.__decode__(data["shantenInfo"]))
        )


class UnionShantenResult(CommonShantenResult[HandPattern]):
    regular: ShantenResult
    chitoi: Optional["ShantenResult"]
    kokushi: Optional["ShantenResult"]

    def __encode__(self) -> dict:
        return dict(
            type='UnionShantenResult',
            hand=self.hand.__encode__(),
            shantenInfo=self.shanten_info.__encode__(),
            regular=self.regular.__encode__(),
            chitoi=self.chitoi.__encode__() if self.chitoi is not None else None,
            kokushi=self.kokushi.__encode__() if self.kokushi is not None else None,
        )

    @classmethod
    def __decode__(cls, data: dict) -> Self:
        return UnionShantenResult(
            hand=Hand.__decode__(data["hand"]),
            shanten_info=cast(CommonShanten, Shanten.__decode__(data["shantenInfo"])),
            regular=RegularShantenResult.__decode__(data["regular"]),
            chitoi=ChitoiShantenResult.__decode__(data["chitoi"]) if data["chitoi"] is not None else None,
            kokushi=KokushiShantenResult.__decode__(data["kokushi"]) if data["kokushi"] is not None else None,
        )


class FuroChanceShantenResult(ShantenResult[ShantenWithFuroChance, RegularHandPattern]):
    def __encode__(self) -> dict:
        return dict(
            type='FuroChanceShantenResult',
            hand=self.hand.__encode__(),
            shantenInfo=self.shanten_info.__encode__()
        )

    @classmethod
    def __decode__(cls, data: dict) -> Self:
        return FuroChanceShantenResult(
            hand=Hand.__decode__(data["hand"], RegularHandPattern),
            shanten_info=ShantenWithFuroChance.__decode__(data["shantenInfo"])
        )



import json
import os
import socket
import sys
from pathlib import Path
from random import randint
from shutil import which
from subprocess import Popen, DEVNULL
from threading import Lock
from time import sleep
from typing import Optional
from urllib.error import HTTPError
from urllib.request import Request, urlopen

from ..protocol import MahjongUtilsBridge
from .path import mahjongutils_webapi_jar_path


def _java_executable() -> Path:
    java_home = os.getenv("JAVA_HOME")
    if java_home is not None:
        java_home = Path(os.getenv("JAVA_HOME")).absolute()
        if sys.platform == 'win32':
            java_executable = java_home / "bin" / "java.exe"  # windows
        else:
            java_executable = java_home / "bin" / "java"  # macos/unix/linux

        if not java_executable.exists():
            raise RuntimeError("Environment variable JAVA_HOME was set to an invalid path. ")
        return java_executable
    else:
        java_executable = which("java")
        if java_executable is not None:
            return Path(java_executable)
        else:
            raise RuntimeError("Cannot find java executable in your PATH environment variable. ")

def _is_port_occupied(port: int):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    code = s.connect_ex(("127.0.0.1", port))
    if code == 0:
        s.close()
        return True
    else:
        return False


class WebApiJarMahjongUtils(MahjongUtilsBridge):
    _process: Optional[Popen]
    _port: Optional[int]

    def __init__(self):
        self._process = None
        self._port = None
        self._init_lock = Lock()
        self.verbose = os.getenv("VERBOSE")

        if self.verbose is None:
            self.verbose = False
        else:
            self.verbose = self.verbose.lower() == "true"

        self._start_process()

    def _choose_port(self) -> int:
        retry = 10
        while retry > 0:
            if self._port is None or _is_port_occupied(self._port):
                self._port = randint(10000, 65535)
                retry -= 1
            else:
                return self._port
        raise RuntimeError(f"Failed to choose a port for some reason. ")

    def _start_process(self):
        if self._process is None or self._process.poll() is not None:
            with self._init_lock:
                if self._process is None or self._process.poll() is not None:
                    port = self._choose_port()
                    self._process = Popen([_java_executable(), "-jar", mahjongutils_webapi_jar_path()],
                                          stdout=sys.stdout if self.verbose else DEVNULL,
                                          stderr=sys.stderr,
                                          env={"PORT": str(port)})

                    handshake_retry = 10
                    while self._process.poll() is None and handshake_retry > 0:
                        if _is_port_occupied(port):
                            return
                        sleep(0.2)
                        handshake_retry -= 1

                    self._process.kill()
                    self._process = None
                    raise RuntimeError(f"Failed to connect with local server at port {port}")

    def call(self, name: str, params: dict) -> dict:
        body = json.dumps(params).encode("utf-8")
        req = Request(f'http://127.0.0.1:{self._port}/{name}', data=body, method="POST")
        try:
            resp = urlopen(req)
            result = resp.read().decode('utf-8')
            result = json.loads(result)
            return result
        except HTTPError as e:
            if e.code == 404 or e.code == 400:
                msg = e.read().decode('utf-8')
                raise ValueError(msg)
            else:
                msg = e.read().decode('utf-8')
                raise RuntimeError(msg)

    def close(self):
        current: Optional[Popen] = getattr(self._process, "value", None)
        current.terminate()



from mahjong_utils.hora import build_hora, build_hora_from_shanten_result
from mahjong_utils.models.furo import Furo
from mahjong_utils.models.tile import Tile, parse_tiles
from mahjong_utils.models.wind import Wind
from mahjong_utils.point_by_han_hu import ParentPoint, ChildPoint
from mahjong_utils.shanten import shanten
from mahjong_utils.yaku.common import ittsu, chinitsu, ipe, tsumo, pinhu, honitsu, sananko, toitoi, self_wind, \
    round_wind, haku, chitoi
from mahjong_utils.yaku.extra import richi, tenhou, ippatsu
from mahjong_utils.yaku.yakuman import churen, tsuiso, daisushi, suanko_tanki, sukantsu, lyuiso, \
    kokushi_thirteen_waiting


def test_build_hora():
    hora = build_hora(parse_tiles("11123456778899p"), None, Tile.by_text("4p"), True, dora=4, extra_yaku={richi})

    assert hora.yaku == {richi, ittsu, chinitsu, ipe, tsumo, pinhu}
    assert hora.han == 16
    assert hora.hu == 20
    assert hora.parent_point == ParentPoint(0, 16000)
    assert hora.child_point == (0, 16000, 8000)


def test_build_hora_2():
    hora = build_hora(parse_tiles("11123456789999p"), None, Tile.by_text("4p"), True, extra_yaku={richi})

    assert hora.yaku == {churen}
    assert hora.han == 13
    assert hora.parent_point == ParentPoint(0, 16000)
    assert hora.child_point == ChildPoint(0, 16000, 8000)


def test_build_hora_3():
    hora = build_hora(parse_tiles("1345556m111z2m"), [Furo.parse("789m")], Tile.by_text("2m"), True)

    assert hora.yaku == {ittsu, honitsu}
    assert hora.han == 3
    assert hora.hu == 40
    assert hora.parent_point == ParentPoint(0, 2600)
    assert hora.child_point == ChildPoint(0, 2600, 1300)


def test_build_hora_4():
    hora = build_hora(parse_tiles("12323467m11z5m"), [Furo.parse("789p")], Tile.by_text("5m"), True, dora=13)

    assert len(hora.yaku) == 0
    assert hora.han == 0
    assert hora.hu == 30
    assert hora.parent_point == ParentPoint(0, 0)
    assert hora.child_point == ChildPoint(0, 0, 0)


def test_build_hora_5():
    hora = build_hora(parse_tiles("11122233344455z"), None, Tile.by_text("5z"), True, extra_yaku={tenhou})

    assert hora.yaku == {tsuiso, daisushi, suanko_tanki, tenhou}
    assert hora.han == 13 * 6
    assert hora.parent_point == ParentPoint(0, 16000 * 6)
    assert hora.child_point == ChildPoint(0, 16000 * 6, 8000 * 6)


def test_build_hora_6():
    hora = build_hora(parse_tiles("66z"),
                      [Furo.parse("3333s"), Furo.parse("2222s"), Furo.parse("0440s"), Furo.parse("8888s")],
                      Tile.by_text("6z"),
                      True)

    assert hora.yaku == {lyuiso, sukantsu}
    assert hora.han == 13 * 2
    assert hora.parent_point == ParentPoint(0, 16000 * 2)
    assert hora.child_point == ChildPoint(0, 16000 * 2, 8000 * 2)


def test_build_hora_7():
    hora = build_hora(parse_tiles("66z"),
                      [Furo.parse("0330s"), Furo.parse("0220s"), Furo.parse("0440s"), Furo.parse("0880s")],
                      Tile.by_text("6z"),
                      True)

    assert hora.yaku == {lyuiso, sukantsu, suanko_tanki}
    assert hora.han == 13 * 4
    assert hora.parent_point == ParentPoint(0, 16000 * 4)
    assert hora.child_point == ChildPoint(0, 16000 * 4, 8000 * 4)


def test_build_hora_8():
    hora = build_hora(parse_tiles("111333555z222s44p"), None, Tile.by_text("2s"), False, dora=4,
                      self_wind=Wind.west, round_wind=Wind.east,
                      extra_yaku={richi, ippatsu})

    assert hora.yaku == {sananko, toitoi, richi, ippatsu, self_wind, round_wind, haku}
    assert hora.han == 13
    assert hora.hu == 60
    assert hora.parent_point == ParentPoint(48000, 0)
    assert hora.child_point == ChildPoint(32000, 0, 0)


def test_build_hora_9():
    hora = build_hora(parse_tiles("55z"),
                      [Furo.parse("0110z"), Furo.parse("0220z"), Furo.parse("0330z"), Furo.parse("0440z")],
                      Tile.by_text("5z"),
                      True)

    assert hora.yaku == {suanko_tanki, sukantsu, daisushi, tsuiso}
    assert hora.han == 13 * 6
    assert hora.parent_point == ParentPoint(0, 16000 * 6)
    assert hora.child_point == ChildPoint(0, 16000 * 6, 8000 * 6)


def test_build_hora_10():
    hora = build_hora(parse_tiles("19s19p19m12345677z"),
                      [],
                      Tile.by_text("7z"),
                      True)

    assert hora.yaku == {kokushi_thirteen_waiting}
    assert hora.han == 13 * 2
    assert hora.parent_point == ParentPoint(0, 16000 * 2)
    assert hora.child_point == ChildPoint(0, 16000 * 2, 8000 * 2)


def test_build_hora_11():
    hora = build_hora(parse_tiles("114477s225588m33z"),
                      [],
                      Tile.by_text("3z"),
                      False)

    assert hora.yaku == {chitoi}
    assert hora.han == 2
    assert hora.parent_point == ParentPoint(2400, 0)
    assert hora.child_point == ChildPoint(1600, 0, 0)


def test_build_hora_from_shanten_result():
    shanten_result = shanten(parse_tiles("11123445678999m"))
    hora = build_hora_from_shanten_result(shanten_result, Tile.by_text("9m"), True)

    assert hora.yaku == {churen}
    assert hora.han == 13
    assert hora.parent_point == ParentPoint(0, 16000)
    assert hora.child_point == ChildPoint(0, 16000, 8000)



from typing import Optional, Sequence

from mahjong_utils.bridge import bridge_mahjongutils
from mahjong_utils.models.furo import Furo
from mahjong_utils.models.tile import Tile
from mahjong_utils.shanten.models import ShantenResult, UnionShantenResult, KokushiShantenResult, ChitoiShantenResult, \
    RegularShantenResult, FuroChanceShantenResult


def regular_shanten(
        tiles: Sequence[Tile],
        furo: Optional[Sequence[Furo]] = None,
        best_shanten_only: bool = False
) -> RegularShantenResult:
    """
    标准形向听分析

    :param tiles: 门前的牌
    :param furo: 副露（对向听分析本身无用，但若需要将结果用于和了分析则需要传入）
    :param best_shanten_only: 仅计算最优向听数的打法（不计算退向打法）
    :return 向听分析结果
    """
    result = bridge_mahjongutils.call("regularShanten", {
        "tiles": [str(t) for t in tiles],
        "furo": [fr.__encode__() for fr in furo] if furo is not None else [],
        "bestShantenOnly": best_shanten_only
    })

    return RegularShantenResult.__decode__(result)


def chitoi_shanten(
        tiles: Sequence[Tile],
        best_shanten_only: bool = False,
) -> ChitoiShantenResult:
    """
    七对子向听分析

    :param tiles: 门前的牌
    :param best_shanten_only: 仅计算最优向听数的打法（不计算退向打法）
    :return 向听分析结果
    """
    result = bridge_mahjongutils.call("chitoiShanten", {
        "tiles": [str(t) for t in tiles],
        "bestShantenOnly": best_shanten_only,
    })

    return ChitoiShantenResult.__decode__(result)


def kokushi_shanten(
        tiles: Sequence[Tile],
        best_shanten_only: bool = False,
) -> KokushiShantenResult:
    """
    国士无双向听分析

    :param tiles: 门前的牌
    :param best_shanten_only: 仅计算最优向听数的打法（不计算退向打法）
    :return 向听分析结果
    """
    result = bridge_mahjongutils.call("kokushiShanten", {
        "tiles": [str(t) for t in tiles],
        "bestShantenOnly": best_shanten_only,
    })

    return KokushiShantenResult.__decode__(result)


def shanten(
        tiles: Sequence[Tile],
        furo: Optional[Sequence[Furo]] = None,
        best_shanten_only: bool = False
) -> UnionShantenResult:
    """
    向听分析

    :param tiles: 门前的牌
    :param furo: 副露（对向听分析本身无用，但若需要将结果用于和了分析则需要传入）
    :param best_shanten_only: 仅计算最优向听数的打法（不计算退向打法）
    :return 向听分析结果
    """
    result = bridge_mahjongutils.call("shanten", {
        "tiles": [str(t) for t in tiles],
        "furo": [fr.__encode__() for fr in furo] if furo is not None else [],
        "bestShantenOnly": best_shanten_only
    })

    return UnionShantenResult.__decode__(result)


def furo_chance_shanten(
        tiles: Sequence[Tile],
        chance_tile: Tile,
        allow_chi: bool = True,
        best_shanten_only: bool = False,
        allow_kuikae: bool = False
):
    """
    副露判断分析

    :param tiles: 门前的牌
    :param chance_tile: 副露机会牌（能够吃、碰的牌）
    :param allow_chi: 是否允许吃
    :param best_shanten_only: 仅计算最优向听数的打法（不计算退向打法）
    :param allow_kuikae: 是否允许食替
    :return 向听分析结果
    """
    result = bridge_mahjongutils.call("furoChanceShanten", {
        "tiles": [str(t) for t in tiles],
        "chanceTile": chance_tile.__encode__(),
        "allowChi": allow_chi,
        "bestShantenOnly": best_shanten_only,
        "allowKuikae": allow_kuikae
    })

    return FuroChanceShantenResult.__decode__(result)


__all__ = ("regular_shanten",
           "chitoi_shanten",
           "kokushi_shanten",
           "furo_chance_shanten",
           "shanten",
           "ShantenResult",)



from enum import Enum

from mahjong_utils.models.tile import Tile
from mahjong_utils.models.tile_type import TileType


class Wind(int, Enum):
    """
    风
    """

    east = 0
    south = 1
    west = 2
    north = 3

    @property
    def tile(self) -> Tile:
        return Tile.by_type_and_num(TileType.Z, self + 1)


__all__ = ("Wind",)



from mahjong_utils.bridge import bridge_mahjongutils


def test_bridge():
    result = bridge_mahjongutils.call("regularShanten", {"tiles": ["1m", "1m", "1m", "2m"]})
    print(result)

    assert result['shantenInfo']['type'] == 'ShantenWithoutGot'
    assert result['shantenInfo']['shantenNum'] == 0



import json
import sys
import threading
from importlib import resources

import cffi

from ..protocol import MahjongUtilsBridge


class LibMahjongUtils(MahjongUtilsBridge):
    def __init__(self) -> None:
        self.ffi = cffi.FFI()

        with resources.open_text(__name__, "libmahjongutils_api.i") as f:
            h = f.read()
            self.ffi.cdef(h)

        if sys.platform == 'win32':
            libname = "libmahjongutils.dll"  # windows
        elif sys.platform == 'darwin':
            libname = "libmahjongutils.dylib"  # macOS
        else:
            libname = "libmahjongutils.so"  # unix/linux

        with resources.path(__name__, libname) as libpath:
            self.lib = self.ffi.dlopen(str(libpath))

        self._lib_sy = threading.local()

    @property
    def lib_sy(self):
        if not hasattr(self._lib_sy, "value"):
            self._lib_sy.value = self.lib.libmahjongutils_symbols()
        return self._lib_sy.value

    def call(self, name: str, params: dict) -> dict:
        params = json.dumps(params)

        result = self.lib_sy.kotlin.root.mahjongutils.entry.call(
            self.ffi.new("char[]", name.encode()),
            self.ffi.new("char[]", params.encode()))
        result = self.ffi.string(result)

        result = json.loads(result)

        if result['code'] == 200:
            return result['data']
        elif result['code'] == 404:
            raise ValueError(result['msg'])
        elif result['code'] == 400:
            raise ValueError(result['msg'])
        else:
            raise RuntimeError(result['msg'])

    def close(self):
        self.ffi.dlclose(self.lib)


__all__ = ("LibMahjongUtils",)



from typing import Set

from .yaku import Yaku
from .extra import tenhou, chihou

kokushi = Yaku("kokushi", 13, 13, True)
suanko = Yaku("suanko", 13, 13, True)
daisangen = Yaku("daisangen", 13, 0, True)
tsuiso = Yaku("tsuiso", 13, 0, True)
shousushi = Yaku("shousushi", 13, 0, True)
lyuiso = Yaku("lyuiso", 13, 0, True)
chinroto = Yaku("chinroto", 13, 0, True)
sukantsu = Yaku("sukantsu", 13, 0, True)
churen = Yaku("churen", 13, 13, True)
daisushi = Yaku("daisushi", 26, 0, True)
churen_nine_waiting = Yaku("churen_nine_waiting", 26, 26, True)
suanko_tanki = Yaku("suanko_tanki", 26, 26, True)
kokushi_thirteen_waiting = Yaku("kokushi_thirteen_waiting", 26, 26, True)

all_yakuman: Set[Yaku] = {
    tenhou, chihou,
    kokushi, suanko, daisangen, tsuiso, shousushi, lyuiso, chinroto, sukantsu, churen,
    daisushi, churen_nine_waiting, suanko_tanki, kokushi_thirteen_waiting
}

__all__ = (
    "all_yakuman",
    "tenhou", "chihou",
    "kokushi", "suanko", "daisangen", "tsuiso", "shousushi", "lyuiso", "chinroto", "sukantsu", "churen",
    "daisushi", "churen_nine_waiting", "suanko_tanki", "kokushi_thirteen_waiting"
)



from typing import FrozenSet

from .yaku import Yaku

tsumo = Yaku("tsumo", 1, 1)
pinhu = Yaku("pinhu", 1, 1)
tanyao = Yaku("tanyao", 1, 0)
ipe = Yaku("ipe", 1, 1)
self_wind = Yaku("self_wind", 1, 0)
round_wind = Yaku("round_wind", 1, 0)
haku = Yaku("haku", 1, 0)
hatsu = Yaku("hatsu", 1, 0)
chun = Yaku("chun", 1, 0)
sanshoku = Yaku("sanshoku", 2, 1)
ittsu = Yaku("ittsu", 2, 1)
chanta = Yaku("chanta", 2, 1)
chitoi = Yaku("chitoi", 2, 2)
toitoi = Yaku("toitoi", 2, 0)
sananko = Yaku("sananko", 2, 0)
honroto = Yaku("honroto", 2, 0)
sandoko = Yaku("sandoko", 2, 0)
sankantsu = Yaku("sankantsu", 2, 0)
shosangen = Yaku("shosangen", 2, 0)
honitsu = Yaku("honitsu", 3, 1)
junchan = Yaku("junchan", 3, 1)
ryanpe = Yaku("ryanpe", 3, 3)
chinitsu = Yaku("chinitsu", 6, 1)

all_common_yaku: FrozenSet[Yaku] = frozenset({
    tsumo, pinhu, tanyao, ipe, self_wind, round_wind, haku, hatsu, chun,
    sanshoku, ittsu, chanta, chitoi, toitoi, sananko, honroto, sandoko, sankantsu, shosangen,
    honitsu, junchan, ryanpe, chinitsu
})

__all__ = (
    "all_common_yaku",
    "tsumo", "pinhu", "tanyao", "ipe", "self_wind", "round_wind", "haku", "hatsu", "chun",
    "sanshoku", "ittsu", "chanta", "chitoi", "toitoi", "sananko", "honroto", "sandoko", "sankantsu", "shosangen",
    "honitsu", "junchan", "ryanpe", "chinitsu"
)



from pathlib import Path

_jar_path: Path = Path("mahjong-utils-webapi-all.jar").absolute()


def mahjongutils_webapi_jar_path() -> Path:
    return _jar_path


def set_mahjongutils_webapi_jar_path(jar_path: Path):
    global _jar_path
    _jar_path = jar_path



import pytest

from mahjong_utils.point_by_han_hu import get_parent_point_by_han_hu, get_child_point_by_han_hu


def test_get_parent_point_by_han_hu():
    assert get_parent_point_by_han_hu(2, 30) == (2900, 1000)
    assert get_parent_point_by_han_hu(4, 40) == (12000, 4000)
    assert get_parent_point_by_han_hu(5, 40) == (12000, 4000)
    assert get_parent_point_by_han_hu(16, 40) == (48000, 16000)

    with pytest.raises(ValueError):
        get_parent_point_by_han_hu(-1, 20)

    with pytest.raises(ValueError):
        get_parent_point_by_han_hu(3, 23)

    with pytest.raises(ValueError):
        get_parent_point_by_han_hu(114, 514)


def test_get_child_point_by_han_hu():
    assert get_child_point_by_han_hu(2, 30) == (2000, 1000, 500)
    assert get_child_point_by_han_hu(4, 40) == (8000, 4000, 2000)
    assert get_child_point_by_han_hu(5, 40) == (8000, 4000, 2000)
    assert get_child_point_by_han_hu(16, 40) == (32000, 16000, 8000)

    with pytest.raises(ValueError):
        get_child_point_by_han_hu(-1, 20)

    with pytest.raises(ValueError):
        get_child_point_by_han_hu(3, 23)

    with pytest.raises(ValueError):
        get_child_point_by_han_hu(114, 514)



from mahjong_utils.models.tile import parse_tiles, all_yaochu, Tile
from mahjong_utils.shanten import shanten, kokushi_shanten, regular_shanten, furo_chance_shanten


def shanten_tester(tiles, expected_shanten,
                   expected_advance,
                   expected_advance_num,
                   expected_good_shape_advance_num=None,
                   expected_improvement_num=None,
                   *,
                   test_func=shanten):
    tiles = parse_tiles(tiles)
    result = test_func(tiles)

    assert result.shanten == expected_shanten
    assert result.advance == expected_advance
    assert result.advance_num == expected_advance_num
    assert result.good_shape_advance_num == expected_good_shape_advance_num
    assert result.improvement_num == expected_improvement_num
    assert result.discard_to_advance is None


# noinspection PyTypeChecker
def test_regular_shanten():
    shanten_tester("34568m235p68s", 2, {*parse_tiles("3678m12345p678s"), }, 40, test_func=regular_shanten)
    shanten_tester("112233p44556s12z", 1, {*parse_tiles("36s12z"), }, 13, 6, test_func=regular_shanten)
    shanten_tester("1112345678999p", 0, {*parse_tiles("123456789p"), }, 23, None, 0, test_func=regular_shanten)
    shanten_tester("114514p1919810s", 2, {*parse_tiles("234567p3456789s"), }, 45, test_func=regular_shanten)


# noinspection PyTypeChecker
def test_shanten():
    shanten_tester("34568m235p68s", 2, {*parse_tiles("3678m12345p678s"), }, 40)
    shanten_tester("3344z6699p11345s", 1, {*parse_tiles("345s"), }, 9, 0)
    shanten_tester("112233p44556s12z", 1, {*parse_tiles("36s12z"), }, 13, 6)
    shanten_tester("1112345678999p", 0, {*parse_tiles("123456789p"), }, 23, None, 0)
    shanten_tester("114514p1919810s", 2, {*parse_tiles("234567p3456789s"), }, 45)
    shanten_tester("119m19p19266s135z", 3, {*parse_tiles("2467z"), }, 16)
    shanten_tester("19m19p19266s1235z", 3, all_yaochu, 42)
    shanten_tester("1119m19p19s12355z", 2, {*parse_tiles("467z"), }, 12)
    shanten_tester("119m19p19s123456z", 0, {*parse_tiles("7z"), }, 4, None, 0)


def shanten_with_got_tester(tiles, expected_shanten, expected_discard_to_advance, test_func=shanten):
    tiles = parse_tiles(tiles)
    result = test_func(tiles)

    actual_discard_to_advance = dict()
    for k, v in result.discard_to_advance.items():
        if v.shanten == result.shanten:
            actual_discard_to_advance[k] = v.advance

    assert result.shanten == expected_shanten
    assert result.advance is None
    assert actual_discard_to_advance == expected_discard_to_advance


# noinspection PyTypeChecker
def test_regular_shanten_with_got_tile():
    shanten_with_got_tester("34568m235p368s", 2, {
        Tile.by_text("5p"): {*parse_tiles("3678m1234p3678s"), },
        Tile.by_text("3s"): {*parse_tiles("3678m12345p678s"), },
        Tile.by_text("2p"): {*parse_tiles("3678m345p3678s"), },
        Tile.by_text("8m"): {*parse_tiles("36m1245p37s"), },
        Tile.by_text("3m"): {*parse_tiles("8m1245p37s"), },
        Tile.by_text("6m"): {*parse_tiles("8m1245p37s"), },
        Tile.by_text("6s"): {*parse_tiles("7m1245p38s"), },
        Tile.by_text("8s"): {*parse_tiles("7m1245p36s"), },
        Tile.by_text("3p"): {*parse_tiles("7m25p37s"), },
    }, regular_shanten)

    shanten_with_got_tester("112233p44556s127z", 1, {
        Tile.by_text("1z"): {*parse_tiles("36s27z"), },
        Tile.by_text("2z"): {*parse_tiles("36s17z"), },
        Tile.by_text("7z"): {*parse_tiles("36s12z"), },
    }, regular_shanten)

    shanten_with_got_tester("11123456789999p", -1, {}, regular_shanten)

    shanten_with_got_tester("1112345678999s1z", 0, {
        Tile.by_text("1z"): {*parse_tiles("123456789s")},
        Tile.by_text("2s"): {*parse_tiles("1z")},
        Tile.by_text("5s"): {*parse_tiles("1z")},
        Tile.by_text("8s"): {*parse_tiles("1z")}
    }, regular_shanten)

    shanten_with_got_tester("114514p1919810s8p", 2, {
        Tile.by_text("8s"): {*parse_tiles("2p3p4p5p6p7p8p9p3s4s5s6s7s9s")},
        Tile.by_text("8p"): {*parse_tiles("2p3p4p5p6p7p3s4s5s6s7s8s9s")},
        Tile.by_text("4p"): {*parse_tiles("3p6p7p8p9p3s4s5s6s7s8s9s")},
        Tile.by_text("9s"): {*parse_tiles("3p4p5p6p7p8p9p3s4s5s6s7s")},
        Tile.by_text("5p"): {*parse_tiles("4p6p7p8p9p3s4s5s6s7s8s9s")},
        Tile.by_text("5s"): {*parse_tiles("2p3p4p5p6p7p8p9p6s7s8s9s")},
    }, regular_shanten)

    shanten_with_got_tester("111p456p123s678p23s", 0, {
        Tile.by_text("1p"): {*parse_tiles("1s4s")},
        Tile.by_text("1s"): {*parse_tiles("23s")},
        Tile.by_text("2s"): {*parse_tiles("3s")},
        Tile.by_text("3s"): {*parse_tiles("2s")}
    }, regular_shanten)

    shanten_with_got_tester("4456m3334556p345s", 0, {
        Tile.by_text("4m"): {*parse_tiles("457p")},
        Tile.by_text("5p"): {*parse_tiles("47m")}
    }, regular_shanten)


# noinspection PyTypeChecker
def test_shanten_with_got_tile():
    shanten_with_got_tester("34568m235p368s", 2, {
        Tile.by_text("5p"): {*parse_tiles("3678m1234p3678s"), },
        Tile.by_text("3s"): {*parse_tiles("3678m12345p678s"), },
        Tile.by_text("2p"): {*parse_tiles("3678m345p3678s"), },
        Tile.by_text("8m"): {*parse_tiles("36m1245p37s"), },
        Tile.by_text("3m"): {*parse_tiles("8m1245p37s"), },
        Tile.by_text("6m"): {*parse_tiles("8m1245p37s"), },
        Tile.by_text("6s"): {*parse_tiles("7m1245p38s"), },
        Tile.by_text("8s"): {*parse_tiles("7m1245p36s"), },
        Tile.by_text("3p"): {*parse_tiles("7m25p37s"), },
    })

    shanten_with_got_tester("3344z6699p11345s8m", 1, {
        Tile.by_text("8m"): {*parse_tiles("345s"), },
        Tile.by_text("3s"): {*parse_tiles("8m45s"), },
        Tile.by_text("4s"): {*parse_tiles("8m35s"), },
        Tile.by_text("5s"): {*parse_tiles("8m34s"), },
    })

    shanten_with_got_tester("112233p44556s127z", 1, {
        Tile.by_text("1z"): {*parse_tiles("36s27z"), },
        Tile.by_text("2z"): {*parse_tiles("36s17z"), },
        Tile.by_text("7z"): {*parse_tiles("36s12z"), },
        Tile.by_text("6s"): {*parse_tiles("127z"), }
    })

    shanten_with_got_tester("11123456789999p", -1, {})

    shanten_with_got_tester("11223344556677z", -1, {})

    shanten_with_got_tester("1112345678999s1z", 0, {
        Tile.by_text("1z"): {*parse_tiles("123456789s")},
        Tile.by_text("2s"): {*parse_tiles("1z")},
        Tile.by_text("5s"): {*parse_tiles("1z")},
        Tile.by_text("8s"): {*parse_tiles("1z")}
    })

    shanten_with_got_tester("114514p1919810s8p", 2, {
        Tile.by_text("8s"): {*parse_tiles("2p3p4p5p6p7p8p9p3s4s5s6s7s9s")},
        Tile.by_text("8p"): {*parse_tiles("2p3p4p5p6p7p3s4s5s6s7s8s9s")},
        Tile.by_text("4p"): {*parse_tiles("3p6p7p8p9p3s4s5s6s7s8s9s")},
        Tile.by_text("9s"): {*parse_tiles("3p4p5p6p7p8p9p3s4s5s6s7s")},
        Tile.by_text("5p"): {*parse_tiles("4p6p7p8p9p3s4s5s6s7s8s9s")},
        Tile.by_text("5s"): {*parse_tiles("2p3p4p5p6p7p8p9p6s7s8s9s")},
        Tile.by_text("1p"): {*parse_tiles("5p8p5s8s")},
        Tile.by_text("1s"): {*parse_tiles("5p8p5s8s")}
    })

    shanten_with_got_tester("111p456p123s678p23s", 0, {
        Tile.by_text("1p"): {*parse_tiles("1s4s")},
        Tile.by_text("1s"): {*parse_tiles("23s")},
        Tile.by_text("2s"): {*parse_tiles("3s")},
        Tile.by_text("3s"): {*parse_tiles("2s")}
    })

    shanten_with_got_tester("4456m3334556p345s", 0, {
        Tile.by_text("4m"): {*parse_tiles("457p")},
        Tile.by_text("5p"): {*parse_tiles("47m")}
    })

    shanten_with_got_tester("119m19p19266s135z3s", 3, {
        Tile.by_text("2s"): {*parse_tiles("2467z"), },
        Tile.by_text("6s"): {*parse_tiles("2467z"), },
        Tile.by_text("3s"): {*parse_tiles("2467z"), }
    }, kokushi_shanten)

    shanten_with_got_tester("19m19p19266s1235z3s", 3, {
        Tile.by_text("2s"): all_yaochu,
        Tile.by_text("6s"): all_yaochu,
        Tile.by_text("3s"): all_yaochu
    }, kokushi_shanten)

    shanten_with_got_tester("1119m19p19s12355z3s", 2, {
        Tile.by_text("1m"): {*parse_tiles("467z"), },
        Tile.by_text("5z"): {*parse_tiles("467z"), },
        Tile.by_text("3s"): {*parse_tiles("467z"), }
    }, kokushi_shanten)

    shanten_with_got_tester("19m19p19s12345566z", 0, {
        Tile.by_text("5z"): {*parse_tiles("7z"), },
        Tile.by_text("6z"): {*parse_tiles("7z"), }
    })

    shanten_with_got_tester("19m19p19s12345667z", -1, {})

    shanten_with_got_tester("35m11223399p7799s", 0, {
        Tile.by_text("3m"): {*parse_tiles("5m"), },
        Tile.by_text("5m"): {*parse_tiles("3m"), }
    })


def test_furo_chance_shanten():
    result = furo_chance_shanten(parse_tiles("3456778m123457p"), Tile.by_text("7m"))
    print(result)



from typing import Optional, Set, List

from stringcase import pascalcase

from mahjong_utils.bridge import bridge_mahjongutils
from mahjong_utils.hora.models import Hora
from mahjong_utils.models.furo import Furo
from mahjong_utils.models.tile import Tile
from mahjong_utils.models.wind import Wind
from mahjong_utils.shanten import ShantenResult
from mahjong_utils.yaku import Yaku


def build_hora(
        tiles: List[Tile], furo: Optional[List[Furo]], agari: Tile,
        tsumo: bool,
        *, dora: int = 0,
        self_wind: Optional[Wind] = None, round_wind: Optional[Wind] = None,
        extra_yaku: Optional[Set[Yaku]] = None
) -> Hora:
    """
    和牌分析

    :param tiles: 手牌
    :param furo: 副露
    :param agari: 和牌
    :param tsumo: 是否自摸
    :param dora: 宝牌数
    :param self_wind: 自风
    :param round_wind: 场风
    :param extra_yaku: 额外役
    :return: 和牌分析结果
    """
    result = bridge_mahjongutils.call("hora", {
        "tiles": [str(t) for t in tiles],
        "furo": [fr.__encode__() for fr in furo] if furo is not None else [],
        "agari": str(agari),
        "tsumo": tsumo,
        "dora": dora,
        "selfWind": pascalcase(self_wind.name) if self_wind is not None else None,
        "roundWind": pascalcase(round_wind.name) if round_wind is not None else None,
        "extraYaku": [pascalcase(yk.name) for yk in extra_yaku] if extra_yaku is not None else []
    })

    return Hora.__decode__(result)


def build_hora_from_shanten_result(
        shanten_result: ShantenResult,
        agari: Tile,
        tsumo: bool,
        *, dora: int = 0,
        self_wind: Optional[Wind] = None, round_wind: Optional[Wind] = None,
        extra_yaku: Optional[Set[Yaku]] = None
) -> Hora:
    """
    和牌分析（根据向听分析结果）

    :param shanten_result: 向听分析结果
    :param agari: 和牌
    :param tsumo: 是否自摸
    :param dora: 宝牌数
    :param self_wind: 自风
    :param round_wind: 场风
    :param extra_yaku: 额外役
    :return: 和牌分析结果
    """
    result = bridge_mahjongutils.call("hora", {
        "shantenResult": shanten_result.__encode__(),
        "agari": str(agari),
        "tsumo": tsumo,
        "dora": dora,
        "selfWind": pascalcase(self_wind.name) if self_wind is not None else None,
        "roundWind": pascalcase(round_wind.name) if round_wind is not None else None,
        "extraYaku": [pascalcase(yk.name) for yk in extra_yaku] if extra_yaku is not None else []
    })

    return Hora.__decode__(result)


__all__ = ("Hora", "build_hora", "build_hora_from_shanten_result")



from abc import ABC, abstractmethod
from typing import Set, Sequence, Union

from pydantic.dataclasses import dataclass

from .mentsu import Mentsu, Shuntsu, Kotsu
from .tile import Tile, parse_tiles, tiles_text
from .tile_type import TileType


@dataclass(frozen=True)
class Tatsu(ABC):
    """
    搭子
    """
    first: Tile

    @property
    @abstractmethod
    def second(self) -> Tile:
        raise NotImplementedError()

    @property
    @abstractmethod
    def waiting(self) -> Set[Tile]:
        raise NotImplementedError()

    def with_waiting(self, tile: Tile) -> Mentsu:
        raise NotImplementedError()

    def __encode__(self) -> str:
        return str(self)

    @classmethod
    def __decode__(cls, data: str) -> "Tatsu":
        return Tatsu.parse(data)

    @staticmethod
    def parse(t: Union[Sequence[Tile], str]) -> "Tatsu":
        if isinstance(t, str):
            t = parse_tiles(t)

        if len(t) != 2:
            raise ValueError("_tiles must has length of 2")

        first, second = t[0], t[1]

        if first > second:
            first, second = second, first

        if first == second:
            return Toitsu(first)
        else:
            if first.tile_type == TileType.Z or second.tile_type == TileType.Z:
                raise ValueError(f"invalid tiles: {t}")

            if second - first == 1:
                if first.num == 1 or first.num == 8:
                    return Penchan(first)
                else:
                    return Ryanmen(first)
            elif second - first == 2:
                return Kanchan(first)
            else:
                raise ValueError(f"invalid tiles: {first}, {second}")


@dataclass(frozen=True)
class Ryanmen(Tatsu):
    """
    两面
    """

    def __post_init__(self):
        assert self.first.num != 1
        assert self.first.num != 8

    @property
    def second(self) -> Tile:
        return self.first + 1

    @property
    def waiting(self) -> Set[Tile]:
        return {self.first - 1, self.second + 1}

    def with_waiting(self, tile: Tile) -> Mentsu:
        if tile == self.first - 1:
            return Shuntsu(tile)
        elif tile == self.second + 1:
            return Shuntsu(self.first)
        else:
            raise ValueError(f"tile {tile} is not waiting")

    def __repr__(self):
        return f"{self.first.num}{self.second.num}{self.first.tile_type.lower()}"


@dataclass(frozen=True)
class Penchan(Tatsu):
    """
    边张
    """

    def __post_init__(self):
        assert self.first.tile_type != TileType.Z
        assert self.first.num == 1 or self.first.num == 8

    @property
    def second(self) -> Tile:
        return self.first + 1

    @property
    def waiting(self) -> Set[Tile]:
        if self.first.num == 1:
            return {self.first + 2}
        else:
            return {self.first - 1}

    def with_waiting(self, tile: Tile) -> Mentsu:
        if self.first.num == 1 and tile == self.first + 2:
            return Shuntsu(self.first)
        elif self.first.num == 8 and tile == self.first - 1:
            return Shuntsu(tile)
        else:
            raise ValueError(f"tile {tile} is not waiting")

    def __repr__(self):
        return f"{self.first.num}{self.second.num}{self.first.tile_type.lower()}"


@dataclass(frozen=True)
class Kanchan(Tatsu):
    """
    坎张
    """

    def __post_init__(self):
        assert self.first.tile_type != TileType.Z
        assert 1 <= self.first.num <= 7

    @property
    def second(self) -> Tile:
        return self.first + 2

    @property
    def waiting(self) -> Set[Tile]:
        return {self.first + 1}

    def with_waiting(self, tile: Tile) -> Mentsu:
        if tile == self.first + 1:
            return Shuntsu(self.first)
        else:
            raise ValueError(f"tile {tile} is not waiting")

    def __repr__(self):
        return f"{self.first.num}{self.second.num}{self.first.tile_type.lower()}"


@dataclass(frozen=True)
class Toitsu(Tatsu):
    """
    对子
    """

    @property
    def second(self) -> Tile:
        return self.first

    @property
    def waiting(self) -> Set[Tile]:
        return {self.first}

    def with_waiting(self, tile: Tile) -> Mentsu:
        if tile == self.first:
            return Kotsu(tile)
        else:
            raise ValueError(f"tile {tile} is not waiting")

    def __repr__(self):
        return f"{self.first.num}{self.second.num}{self.first.tile_type.lower()}"


__all__ = ("Tatsu", "Toitsu", "Kanchan", "Ryanmen", "Penchan",)
